from typing import Dict, List, Union

import pandas


def _compile_parent_linkage(edges: pandas.DataFrame) -> Dict[str, List[str]]:
	""" Maps a genotype to a list of all genotypes that inherit from it"""
	children = dict()
	for _, row in edges.iterrows():
		parent = row['Parent']
		identity = row['Identity']

		children[parent] = children.get(parent, list()) + [identity]
	return children


def _subtract_children_from_parent(modified_genotypes: pandas.DataFrame, children: Dict[str, List[str]], detection_cutoff: float) -> pandas.DataFrame:
	children_table = list()
	for genotype_label, genotype in modified_genotypes.iterrows():
		if genotype_label in children:
			genotype_frequencies = genotype[genotype > detection_cutoff]
			genotype_children: pandas.Series = modified_genotypes.loc[children[genotype_label]].max()
			genotype_frequencies: pandas.Series = genotype_frequencies - genotype_children
			genotype_frequencies = genotype_frequencies.mask(lambda s: s < detection_cutoff, 0.01)  # 0.05 so there is still a visible slice.
			genotype_frequencies = genotype_frequencies.fillna(0)  # Otherwise plotting the muller diagram will fail.
		else:
			genotype_frequencies = genotype
		genotype_frequencies.name = genotype_label
		children_table.append(genotype_frequencies)
	return pandas.DataFrame(children_table)


def _convert_genotype_table_to_population_table(genotype_table: pandas.DataFrame) -> pandas.DataFrame:
	""" Pivots a genotype table into a long-form table.
		Parameters
		----------
		genotype_table: pandas.DataFrame
			 The genotype table to pivot.

		Returns
		-------
		pandas.DatFrame
			- columns
				`Generation`: int
				`Identity`: str
				`Population`: float
	"""
	table = list()
	for genotype_label, genotype_frequencies in genotype_table.iterrows():
		# Remove timepoints where the value was 0 or less than 0 due to above line.
		for timepoint, frequency in genotype_frequencies.items():
			row = {
				'Identity':   genotype_label,
				'Generation': int(timepoint),
				'Population': frequency * 100
			}
			table.append(row)

	temp_df = pandas.DataFrame(table)
	return temp_df


def _append_genotype_0(population_table: pandas.DataFrame) -> pandas.DataFrame:
	generation_groups = population_table.groupby(by = 'Generation')
	modified_population = population_table
	for generation, group in generation_groups:
		population = group['Population'].sum()
		if population <= 100:
			p = 100 - population
		else:
			p = 0
		modified_population = modified_population.append({'Generation': generation, "Identity": "genotype-0", "Population": p}, ignore_index = True)
	return modified_population


def generate_ggmuller_population_table(mean_genotypes: pandas.DataFrame, edges: pandas.DataFrame, detection_cutoff: float,
		adjust_populations: bool) -> pandas.DataFrame:
	"""
		Converts the genotype frequencies to a population table suitable for ggmuller.
	Parameters
	----------
	mean_genotypes: pandas.DataFrame
		Contains the mean frequency for each genotype at each timepoint.
	edges: pandas.DataFrame
		The output from create_ggmuller_edges()
	detection_cutoff: float
		The cutoff to determine whether a trajectory or genotype counts as being nonzero at a given timepoint.
	adjust_populations:bool
		If true, subtracts any children genotypes from the parent's frequency. Used to make the muller diagrams more intuitive.
	Returns
	-------
	pandas.DataFrame
	"""

	# "Generation", "Identity" and "Population"
	# Adjust populations to account for inheritance.
	# If a child genotype fixed, the parent genotype should be replaced.

	# Use a copy of the dataframe to avoid making changes to the original.
	modified_genotypes: pandas.DataFrame = mean_genotypes.copy(True)
	# In case the genotype table includes genotypes that the edges table does not have.
	modified_genotypes = modified_genotypes[modified_genotypes.index.isin(edges['Identity'])]

	# Generate a list of all genotypes that arise in the background of each genotype.
	# Should ba a dict mapping parent -> list[children]
	children = _compile_parent_linkage(edges)
	if adjust_populations:
		child_df = _subtract_children_from_parent(modified_genotypes, children, detection_cutoff)
	else:
		child_df = modified_genotypes
	temp_df = _convert_genotype_table_to_population_table(child_df)
	population_table = _append_genotype_0(temp_df)

	return population_table


def generate_missing_trajectories_table(trajectories: pandas.DataFrame, original_trajectories: pandas.DataFrame) -> pandas.DataFrame:
	missing_trajectories = original_trajectories[~original_trajectories.index.isin(trajectories.index)]
	concat_trajectories = pandas.concat([trajectories, missing_trajectories], sort = False)

	return concat_trajectories


def generate_trajectory_table(trajectories: pandas.DataFrame, parent_genotypes: Union[pandas.Series, Dict[str, str]],
		info: pandas.DataFrame) -> pandas.DataFrame:
	# Sorts the trajectory table and adds additional columns from the original table.
	trajectories = trajectories[sorted(trajectories.columns, key = lambda s: int(s))]
	if info is not None:
		trajectory_table: pandas.DataFrame = trajectories.copy()
		trajectory_table['genotype'] = [parent_genotypes.get(k) for k in trajectory_table.index]
		trajectory_table = trajectory_table.join(info).sort_values(by = ['genotype'])
		return trajectory_table



import math

import pandas

from muller.widgets import get_valid_points

try:
	from muller.inheritance import order_by_area
except ModuleNotFoundError:
	from . import order_by_area


def calculate_additive_score(nested_genotype: pandas.Series, unnested_genotype: pandas.Series, cutoff: float) -> float:
	"""
		Tests whether the nested genotype is consistenty larger than  the nested genotype. The scoring is as follows:
		`nested_genotype` always larger than `unnested_genotype`: 1
		`nested_genotype` consistently larger than `unnested_genotype`: 0.5
		`nested_genotype` not consistently larger than `unnested_genotype`: 0
		`nested_genotype` sometimes smaller than `unnested_genotype`: -0.5
	Parameters
	----------
	nested_genotype, unnested_genotype: pandas.Series
	cutoff: The minimum distance between each timepoint between `nested_genotype` and `unnested_genotype`

	"""
	difference = nested_genotype - unnested_genotype

	score = int(difference.sum() > (cutoff * len(difference) / 2))

	# There may be a point where the unnested genotype is greater than the nested genotype.

	if any(difference < 0):
		score -= 1
	# logger.debug(f"{unnested_genotype.name}\t{nested_genotype.name}\t{score}")
	return score


def calculate_derivative_score(left: pandas.Series, right: pandas.Series, detection_cutoff: float, cutoff: float) -> float:
	"""
		Tests whther the two series are correlated or anticorrelated with each other. The scoring is as follows:
		correlated: 2
		uncorrelated: 0
		anticorrelated: -2
	Parameters
	----------
	left, right: pandas.Series
		The two series to test.
	detection_cutoff: float
		Used to determine which points to use for the comparison.
	cutoff: float
		The cutoff value to determine whther the series are actually correlated/anticorrelated. Any value within the range [-`cutoff`, `cutoff`]
		results in a score of 0.
	"""
	# Pandas implementation of the derivative check, since it basically just checks for covariance.
	valid_left, valid_right = get_valid_points(left, right, detection_cutoff, flimit = 0.97)
	if valid_left.empty:
		covariance = math.nan
	else:
		covariance = valid_left.cov(valid_right)

	if covariance > cutoff: score = 2
	elif covariance < -cutoff: score = -2
	else: score = 0
	from loguru import logger
	logger.debug(f"{right.name}\t{left.name}\t{score}\t{covariance}")
	return score


def calculate_area_score(nested_genotype: pandas.Series, unnested_genotype: pandas.Series) -> float:
	"""
		Calculates a score based on the probability the unnested genotype is a subset of the nested_genotype.
		This take into account both the area of the nested genotype and the area of all other genotypes not in the nested genotype.
	Parameters
	----------
	nested_genotype, unnested_genotype: pandas.Series

	"""

	# If the nested genotype is not fixed, group the remaining frequencies into an `other` category.
	# noinspection PyTypeChecker
	other_genotypes: pandas.Series = 1 - nested_genotype
	area_nested = order_by_area.area_of_series(nested_genotype)
	area_unnested = order_by_area.area_of_series(unnested_genotype)
	area_other = order_by_area.area_of_series(other_genotypes)
	common_area_nested = order_by_area.calculate_common_area(nested_genotype, unnested_genotype)
	common_area_other = order_by_area.calculate_common_area(other_genotypes, unnested_genotype)

	is_subset_nested = order_by_area.is_subset(area_nested, area_unnested, common_area_nested)
	is_subset_other = order_by_area.is_subset(area_other, area_unnested, common_area_other)

	if is_subset_nested and is_subset_other:
		score = int(common_area_nested > 2 * common_area_other)
	elif is_subset_nested:
		score = 2
	else:
		score = -2
	# logger.debug(f"{unnested_genotype.name}\t{nested_genotype.name}\t{score}")
	return score


def calculate_subtractive_score(left: pandas.Series, right: pandas.Series, fixed_cutoff: float, cutoff: float) -> int:
	"""
		Tests whether two genotypes consistently sum to a value greater than the fixed breakpoint. This suggests that one of the genotypes
		is in the background of the other, since otherwise the maximum combined frequency should, at most, be equal to the fixed cutoff value.
	Parameters
	----------
	left, right: pandas.Series
	fixed_cutoff: The fixed breakpoint.
	cutoff: Governs whether two series consistently sum to a value greater than `fixed_cutoff`. Should be in the range [0,1]
	"""

	combined = left + right
	above_fixed = combined - fixed_cutoff

	# Finally, check whether the frequency of this series is consistently above the `cutoff` value.
	# Timepoints that do not sum to greater than the fixed breakpoint will be negative.
	result = above_fixed.mean() > cutoff

	return int(result)



#!python
import muller
if __name__ == "__main__":
	args = muller.commandline_parser.create_parser().parse_args()

	muller.muller_workflow.workflow(args.filename, args.output_folder, program_options = args) 



import json
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, Optional, Tuple

import pandas
from loguru import logger

ROOT_GENOTYPE_LABEL = "genotype-0"
FILTERED_GENOTYPE_LABEL = "removed"
OutputType = Tuple[pandas.DataFrame, pandas.DataFrame, str, Dict[str, Any]]

try:
	from muller.clustering.metrics.pairwise_calculation_cache import PairwiseCalculationCache
	from muller.graphics import plot_genotypes, plot_heatmap, plot_dendrogram, generate_muller_plot, plot_timeseries, flowchart
	from muller.muller_output.generate_tables import *
	from muller.muller_output.generate_scripts import generate_r_script, execute_r_script
	from muller import widgets, dataio, palettes

except ModuleNotFoundError:
	from .. import widgets, dataio, palettes
	from ..clustering.metrics.pairwise_calculation_cache import PairwiseCalculationCache
	from ..graphics import plot_genotypes, plot_heatmap, plot_dendrogram, generate_muller_plot, plot_timeseries, flowchart
	from .generate_tables import *
	from .generate_scripts import generate_r_script, execute_r_script


@dataclass
class WorkflowData:
	# Used to organize the output from the workflow.
	version: str
	filename: Path
	program_options: Any
	info: Optional[pandas.DataFrame]
	original_trajectories: Optional[pandas.DataFrame]
	original_genotypes: Optional[pandas.DataFrame]
	rejected_trajectories: pandas.DataFrame
	trajectories: pandas.DataFrame
	genotypes: pandas.DataFrame
	genotype_members: pandas.Series
	clusters: Any
	p_values: PairwiseCalculationCache
	filter_cache: List[Tuple[pandas.DataFrame, pandas.DataFrame]]
	linkage_matrix: Any
	genotype_palette_filename: Optional[Path]


class OutputFilenames:
	""" Used to organize the files generated by the workflow.
	"""

	def __init__(self, output_folder: Path, name: str, suffix = 'tsv'):
		self.suffix = suffix

		def check_folder(path: Union[str, Path]) -> Path:
			path = Path(path)
			if not path.exists():
				path.mkdir()
			return path.absolute()

		output_folder = check_folder(output_folder)
		supplementary_folder = check_folder(output_folder / "supplementary-files")
		graphics_folder = check_folder(output_folder / "graphics")
		graphics_distinctive_folder = check_folder(graphics_folder / "distinctive")
		graphics_clade_folder = check_folder(graphics_folder / "clade")
		tables_folder = check_folder(output_folder / "tables")
		scripts_folder = check_folder(output_folder / "scripts")
		# General Files
		self.trajectory: Path = output_folder / (name + f'.trajectories.{suffix}')
		self.genotype: Path = output_folder / (name + f'.muller_genotypes.{suffix}')
		self.muller_plot_annotated: Path = output_folder / (name + '.muller.annotated.png')
		self.muller_plot_unannotated_general: Path = output_folder / (name + '.muller.unannotated.png')

		# tables
		self.original_trajectory: Path = tables_folder / (name + f'.trajectories.original.{suffix}')
		self.original_genotype: Path = tables_folder / (name + f'.muller_genotypes.original.{suffix}')
		self.rejected_trajectories: Path = tables_folder / (name + f"trajectories.rejected.{suffix}")
		self.population: Path = tables_folder / (name + f'.ggmuller.populations.{suffix}')
		self.edges: Path = tables_folder / (name + f'.ggmuller.edges.{suffix}')
		self.muller_table: Path = tables_folder / (name + f'.muller.csv')  # This is generated in r.

		self.linkage_matrix_table = tables_folder / (name + f".linkagematrix.tsv")
		self.distance_table: Path = tables_folder / (name + ".distance.tsv")
		self.distance_matrix: Path = tables_folder / (name + f".distance.{suffix}")

		# graphics
		# Muller Plots
		self.muller_plot_unannotated: Path = graphics_distinctive_folder / (name + '.muller.unannotated.png')
		self.muller_plot_annotated_pdf: Path = graphics_clade_folder / (name + '.muller.annotated.pdf')
		self.muller_plot_annotated_svg: Path = graphics_clade_folder / (name + ".muller.annotated.svg")
		self.muller_plot_basic: Path = graphics_clade_folder / (name + '.muller.basic.png')
		self.muller_plot_annotated_distinctive: Path = graphics_distinctive_folder / (name + '.muller.annotated.distinctive.png')
		self.muller_plot_annotated_distinctive_svg: Path = graphics_distinctive_folder / (name + '.muller.annotated.distinctive.svg')

		##Timeseries plots
		self.genotype_plot: Path = graphics_distinctive_folder / (name + '.genotypes.distinctive.png')
		self.trajectory_plot_distinctive: Path = graphics_distinctive_folder / (name + f".trajectories.distinctive.svg")
		self.genotype_plot_filtered: Path = output_folder / (name + f".genotypes.filtered.png")

		## Geneology plots
		self.lineage_render: Path = output_folder / (name + '.lineage.svg')
		self.lineage_image_clade: Path = graphics_distinctive_folder / (name + f".lineage.distinctive.png")
		self.lineage_image_distinct: Path = output_folder / (name + '.lineage.png')

		## Other plots
		self.distance_heatmap: Path = graphics_folder / (name + f".heatmap.distance.png")
		self.linkage_plot = graphics_folder / (name + f".dendrogram.png")

		# scripts
		self.r_script: Path = scripts_folder / (name + '.r')
		self.mermaid_script: Path = scripts_folder / (name + '.mermaid.md')

		# supplementary files
		self.parameters: Path = supplementary_folder / (name + '.json')
		self.genotype_scores: Path = supplementary_folder / (name + '.nestscores.tsv')

	@property
	def delimiter(self) -> str:
		if self.suffix == 'tsv':
			return '\t'
		else:
			return ','


def get_workflow_parameters(workflow_data: WorkflowData, genotype_colors = Dict[str, str]) -> Dict[str, float]:
	parameters = {k: (v if not isinstance(v, Path) else str(v)) for k, v in workflow_data.program_options.items()}
	parameters['genotypeColors'] = genotype_colors
	parameters['version'] = workflow_data.version

	return parameters


def generate_output(workflow_data: WorkflowData, output_folder: Path, detection_cutoff: float, adjust_populations: bool):
	# Set up the output folder
	if workflow_data.program_options['name']:
		base_filename = workflow_data.program_options['name']
	else:
		base_filename = workflow_data.filename.stem
	if workflow_data.program_options['sheetname'] and workflow_data.program_options['sheetname'] != 'Sheet1':
		base_filename += '.' + workflow_data.program_options['sheetname']

	filenames = OutputFilenames(output_folder, base_filename)
	delimiter = filenames.delimiter

	# Map each trajectory to its parent genotype.
	parent_genotypes = widgets.map_trajectories_to_genotype(workflow_data.genotype_members)

	##############################################################################################################################################
	# ------------------------------------------- Save the genotype and trajectory tables --------------------------------------------------------
	##############################################################################################################################################

	workflow_data.genotypes.to_csv(str(filenames.genotype), sep = delimiter)

	logger.info("Saving Trajectory Tables...")
	# Save trajectory tables, if available
	if workflow_data.trajectories is not None:
		filtered_trajectories = generate_missing_trajectories_table(workflow_data.trajectories, workflow_data.original_trajectories)
		trajectories = generate_trajectory_table(workflow_data.trajectories, parent_genotypes, workflow_data.info)
		trajectories.to_csv(str(filenames.trajectory), sep = delimiter)
	if workflow_data.rejected_trajectories is not None:
		workflow_data.rejected_trajectories.to_csv(filenames.rejected_trajectories, sep = delimiter)

	##############################################################################################################################################
	# ----------------------------------------- # Generate the input tables to ggmuller ----------------------------------------------------------
	##############################################################################################################################################
	edges_table = workflow_data.clusters.as_ancestry_table().reset_index()
	edges_table = edges_table[['Parent', 'Identity']]  # Otherwise the r script doesn't work.
	population_table = generate_ggmuller_population_table(workflow_data.genotypes, edges_table, detection_cutoff, adjust_populations)
	population_table.to_csv(str(filenames.population), sep = delimiter, index = False)
	edges_table.to_csv(str(filenames.edges), sep = delimiter, index = False)
	##############################################################################################################################################
	# ----------------------------------------- Generate the palette for the praphics ------------------------------------------------------------
	##############################################################################################################################################
	_all_genotype_labels = sorted(set(list(workflow_data.original_genotypes.index) + list(workflow_data.genotypes.index)))
	# Annotations may be used to select specific colors for the lineage palette.
	if workflow_data.info is not None:
		genotype_annotations = dataio.parse_genotype_annotations(
			workflow_data.genotype_members,
			workflow_data.info,
			workflow_data.program_options['alias_filename']
		)
	else:
		genotype_annotations = {}
	# The custom palette overrides any other color.
	if workflow_data.genotype_palette_filename:
		custom_palette = dataio.read_map(workflow_data.genotype_palette_filename)
	else:
		custom_palette = {}

	genotype_colors_distinct = palettes.generate_palette(_all_genotype_labels)
	genotype_colors_clade = palettes.generate_palette(edges_table, custom_palette, genotype_annotations, 'lineage')

	##############################################################################################################################################
	# ------------------------------------------------- Save supplementary files -----------------------------------------------------------------
	##############################################################################################################################################
	parameters = get_workflow_parameters(workflow_data, genotype_colors_clade)
	filenames.parameters.write_text(json.dumps(parameters, indent = 2))

	# Generate and excecute scripts
	##############################################################################################################################################
	# ------------------------------------------------- Generate the lineage plots ---------------------------------------------------------------
	##############################################################################################################################################
	logger.info("Generating Lineage Plots...")
	edges_table = workflow_data.clusters.priority_table()
	flowchart.flowchart(edges_table, genotype_colors_clade, annotations = genotype_annotations, filename = filenames.lineage_image_distinct)
	flowchart.flowchart(edges_table, genotype_colors_distinct, annotations = genotype_annotations, filename = filenames.lineage_render)
	flowchart.flowchart(edges_table, genotype_colors_distinct, annotations = genotype_annotations, filename = filenames.lineage_image_clade)

	##############################################################################################################################################
	# ------------------------------------------------- Generate and excecute the r script -------------------------------------------------------
	##############################################################################################################################################
	# TODO replicate how the r script generates the muller_df file so that we don't have to run the rscript.
	muller_df = generate_r_script(
		trajectory = filenames.trajectory,
		population = filenames.population,
		edges = filenames.edges,
		table_filename = filenames.muller_table,
		plot_filename = filenames.muller_plot_basic,
		script_filename = filenames.r_script,
		color_palette = genotype_colors_clade,
		genotype_labels = population_table['Identity'].unique().tolist()
	)

	# Generate time series plots showing the mutations/genotypes over time.

	##############################################################################################################################################
	# -------------------------------- Generate time series plots showing the mutations/genotypes over time --------------------------------------
	##############################################################################################################################################
	logger.info("Generating series plots...")
	trajectory_colors_distinct = {k: genotype_colors_distinct[v] for k, v in parent_genotypes.items()}
	trajectory_colors_lineage = {k: genotype_colors_clade[v] for k, v in parent_genotypes.items()}

	plot_genotypes(workflow_data.trajectories, workflow_data.genotypes, filenames.genotype_plot, genotype_colors_distinct, trajectory_colors_distinct)
	if workflow_data.trajectories is not None:
		plot_timeseries(workflow_data.trajectories, trajectory_colors_distinct, filename = filenames.trajectory_plot_distinctive)
		plot_genotypes(filtered_trajectories, workflow_data.genotypes, filenames.genotype_plot_filtered, genotype_colors_clade,
			trajectory_colors_lineage)

	##############################################################################################################################################
	# ------------------------------------- Generate the muller plot using the table from the r script -------------------------------------------
	##############################################################################################################################################
	logger.info("Generating muller plots...")
	if muller_df is not None:
		annotated_muller_plot_filenames = [
			filenames.muller_plot_annotated,
			filenames.muller_plot_annotated_pdf,
			filenames.muller_plot_annotated_svg
		]
		distinctive_muller_plot_filenames = [
			filenames.muller_plot_annotated_distinctive,
			filenames.muller_plot_annotated_distinctive_svg
		]
		use_outlines = not workflow_data.program_options['no_outline']
		generate_muller_plot(muller_df, genotype_colors_clade, annotated_muller_plot_filenames, genotype_annotations, outlines = use_outlines)
		generate_muller_plot(muller_df, genotype_colors_clade, [filenames.muller_plot_unannotated_general], outlines = use_outlines)
		generate_muller_plot(muller_df, genotype_colors_distinct, distinctive_muller_plot_filenames, outlines = use_outlines)

	##############################################################################################################################################
	# ---------------------------------------------------- Generate supplementary files ----------------------------------------------------------
	##############################################################################################################################################
	workflow_data.clusters.to_table().to_csv(filenames.genotype_scores, sep = '\t')
	if workflow_data.linkage_matrix is not None:
		num_trajectories = len(workflow_data.trajectories)
		linkage_table = widgets.format_linkage_matrix(workflow_data.linkage_matrix, num_trajectories)
		linkage_table.to_csv(str(filenames.linkage_matrix_table), sep = delimiter, index = False)
		try:
			plot_dendrogram(workflow_data.linkage_matrix, workflow_data.p_values, filenames.linkage_plot)
		except:
			pass
	workflow_data.p_values.save(filenames.distance_table)

	try:
		squareform = workflow_data.p_values.squareform()
		squareform.to_csv(filenames.distance_matrix)
	except:
		pass

	pvalues_matrix = workflow_data.p_values.squareform()
	try:
		plot_heatmap(pvalues_matrix, filenames.distance_heatmap)
	except:
		pass



from typing import Dict, List, Tuple

import pandas
from loguru import logger

try:
	from muller.inheritance import scoring
	from muller.inheritance.cluster import Cluster
	from muller import widgets
except ModuleNotFoundError:
	from . import scoring
	from .cluster import Cluster
	from .. import widgets


def order_clusters(sorted_df: pandas.DataFrame, dlimit: float, flimit: float, additive_cutoff: float, subtractive_cutoff: float,
		derivative_cutoff: float,
		known_ancestry: Dict[str, str] = None) -> Cluster:
	"""
		Orders genotypes by which background they belong to.
	Parameters
	----------
	sorted_df: pandas.DataFrame
		A dataframe of sorted genotypes based on when the genotype was first detected and first fixed.
	dlimit: float
		The detection limit
	flimit: float
		The cutoff value to consider a genotype "fixed"
	additive_cutoff: float
		Used when testing whther a nested genotype is consistently greater than an unnested genotype
	subtractive_cutoff: float
		Used to test whether the combined frequencies of the nested/unnested genotype are consistently greater than the fixed cutoff.
	derivative_cutoff: float
		Used when testing whether two genotypes are correlated, not correlated, or anticorrelated. correlated/anticorrelated genotypes
		must have a covariance outside the range [-`derivative_cutoff`, `derivative_cutoff`].
	known_ancestry: Dict[str,str]
		Manually-assigned ancestry values. For now, the parent genotype is automatically assigned to the root genotype to prevent
		circular links from forming.

	Returns
	-------
	ClusterType
	"""
	# By default the backgrounds should occupy the first n lines of the dataframe
	if known_ancestry is None: known_ancestry = dict()

	initial_background = sorted_df.iloc[0]
	genotype_nests = Cluster(initial_background, timepoints = sorted_df)
	if known_ancestry:
		logger.info(f"Found user-given ancestries.")
	for identity, parent in known_ancestry.items():
		# TODO need to add a way to prevent circular ancestry links when a user manually assigns ancestry. Current workaround forces the manual parent to be in the root background.
		# TODO Also need to add a way to specify a genotype by one of the linked annotations.
		logger.info(f"Adding {parent} as a potential background for {identity}")
		genotype_nests.add_genotype_to_background(parent, 'genotype-0', priority = 100)
		genotype_nests.add_genotype_to_background(identity, parent, priority = 100)  # Dummy priority so that it is selected before other backgrounds.
	for unnested_label, unnested_trajectory in sorted_df[1:].iterrows():
		logger.debug(f"Nesting {unnested_label}")
		# Iterate over the rest of the table in reverse order. Basically, we start with the newest nest and iterate until we find a nest that satisfies the filters.
		test_table = sorted_df[:unnested_label].iloc[::-1]
		for nested_label, nested_genotype in test_table.iterrows():
			if nested_label == unnested_label: continue
			score_additive = scoring.calculate_additive_score(nested_genotype, unnested_trajectory, additive_cutoff)
			score_subtractive = scoring.calculate_subtractive_score(nested_genotype, unnested_trajectory, flimit, subtractive_cutoff)
			# The derivative score should only be computed using the timepoints where the series overlap.
			detected_left, detected_right = widgets.get_valid_points(nested_genotype, unnested_trajectory, dlimit = dlimit, inner = True)
			score_derivative = scoring.calculate_derivative_score(detected_left, detected_right, detection_cutoff = dlimit,
				cutoff = derivative_cutoff)
			score_area = scoring.calculate_area_score(nested_genotype, unnested_trajectory)

			total_score = score_additive + score_subtractive + score_area
			if total_score > 0:
				# The derivative check is only useful when deciding between possible candidates, since it does not provide evidence itself that a
				# genotype is a potential background. So, at least one of the other checks should have been passed with no
				# evidence against the candidate background.
				total_score += derivative_cutoff
			logger.debug(f"{unnested_label}\t{nested_label}\t{total_score}|{score_additive}\t{score_subtractive}\t{score_derivative}\t{score_area}")
			genotype_nests.add_genotype_to_background(unnested_label, nested_label, total_score)
	logger.debug("Final Ancestry:")
	for genotype_label in sorted_df.index:
		candidate = genotype_nests.get_highest_priority(genotype_label)
		logger.debug(f"{genotype_label}\t{candidate}")
	return genotype_nests


def get_maximum_genotype_delta(genotype_deltas: List[Tuple[str, float]]) -> Tuple[str, float]:
	if genotype_deltas:
		correlated_label, correlated_delta = max(genotype_deltas, key = lambda s: s[1])
	else:
		correlated_label = "N/A"  # Shouldn't end up being used.
		correlated_delta = 0
	return correlated_label, correlated_delta



import math
from typing import Set, Union

import pandas


def area_of_series(series: pandas.Series) -> float:
	""" Calculates the area of a discrete series."""
	total = series.sum()
	return total


def calculate_common_area(left: pandas.Series, right: pandas.Series) -> float:
	""" Calculates |X and Y|"""
	overlap_index = get_overlap(left, right)
	if not overlap_index:
		overlap = 0
	else:
		overlap = area_of_series(left.where(left < right, right))
	return overlap


def calculate_area_difference(left: pandas.Series, right: pandas.Series):
	left_area = area_of_series(left)
	right_area = area_of_series(right)
	return left_area - right_area


# noinspection PyTypeChecker
def get_overlap(left: pandas.Series, right: pandas.Series) -> Set[int]:
	detected_left = left
	detected_right = right

	overlap_index = set(detected_left.index) & set(detected_right.index)
	return overlap_index


def X_or_Y(left: Union[float, pandas.Series], right: Union[float, pandas.Series], common: float = None) -> float:
	""" Calculated the XOR area between `left` and `right`
		Parameters
		----------
		left, right: pandas.Series
			The two series to calculate the exclusive area on.
		left, right,common: float
			Precomputed values for the area of `left`, `right`, and x_and_y.
	"""
	if common is None:
		area_left = area_of_series(left)
		area_right = area_of_series(right)
		common_area = calculate_common_area(left, right)
	else:
		area_left, area_right, common_area = left, right, common

	return area_left + area_right - common_area


def jaccard_distance(left: Union[float, pandas.Series], right: Union[float, pandas.Series]) -> float:
	""" Calulated the jaccard distance between two series.
		Parameters
		----------
		left, right: pandas.Series
			The two genotypes to calculate the jaccard distance on.
		left, right: float
			precomputed x_or_y and x_and_y, respectively.
	"""
	if isinstance(left, float):
		x_or_y, x_and_y = left, right
	else:
		x_or_y = X_or_Y(left, right)
		x_and_y = X_and_Y(left, right)
	return (x_or_y - x_and_y) / x_or_y


def jaccard_subset(left: Union[float, pandas.Series], right: Union[float, pandas.Series]) -> float:
	"""
		Calculates the jaccard distance assuming `right` is a subset of `left`.
	Parameters
	----------
	left, right: pandas.Series
	left, right: float
	"""
	if isinstance(left, float):
		left_area, right_area = left, right
	else:
		left_area = area_of_series(left)
		right_area = area_of_series(right)
	return (left_area - right_area) / left_area


def is_subset_legacy(left: pandas.Series, right: pandas.Series) -> bool:
	""" Tests if `right` is a subset of `left`. Deprecated since it needs to be faster."""
	actual_jaccard_distance = jaccard_distance(left, right)
	subset_jaccard_distance = jaccard_subset(left, right)
	# TODO: incorporate variance into this
	return math.isclose(actual_jaccard_distance, subset_jaccard_distance, abs_tol = 0.1)


def is_subset(left: Union[float, pandas.Series], right: Union[float, pandas.Series], common: float = None):
	"""
		Estimates whether `right` is a subset of `left`. This implementation should be a little faster than the legacy method.
	Parameters
	----------
	left, right: pandas.Series
	left,right,common: float
		precomuted areas
	"""
	if common is None:
		left_area = area_of_series(left)
		right_area = area_of_series(right)
		x_and_y = calculate_common_area(left, right)
	else:
		left_area, right_area, x_and_y = left, right, common
	x_or_y = X_or_Y(left_area, right_area, x_and_y)

	actual_jaccard_distance = jaccard_distance(x_or_y, x_and_y)
	subset_jaccard_distance = jaccard_subset(left_area, right_area)  # Jaccard distance assuming `left` is a subset of `right`
	return math.isclose(actual_jaccard_distance, subset_jaccard_distance, abs_tol = 0.1)


X_and_Y = calculate_common_area



from pathlib import Path
from typing import Any, Optional, Tuple, Union
import math
import pandas
from loguru import logger
import re
try:
	from muller.widgets import get_numeric_columns
	from muller.dataio import import_table
except ModuleNotFoundError:
	from ..widgets import get_numeric_columns
	from . import import_table

IOTYPE = Union[str, Path]


def _convert_to_integer(value: Any, default: Optional[int] = None) -> int:
	""" Attempts to convert the input value to an integer. Returns `default` otherwise."""
	if isinstance(value, str) and (value.startswith('x') or value.startswith('X')):
		value = value[1:]
	try:
		result = int(value)
	except (TypeError, ValueError):
		result = default
	return result


def _correct_math_scale(old_data: pandas.DataFrame) -> pandas.DataFrame:
	""" Ensures the time table columns contain values between 0 and 1 and are of type `float`"""
	new_data = old_data.copy(deep = True)
	for column in old_data.columns:
		try:
			maximum_value = old_data[column].max()
		except TypeError as exception:
			logger.error(f"Some of the values in the column '{column}' could not be read as numbers.")
			logger.error(f"The column had values of {old_data[column].values}")
			raise exception
		if maximum_value > 1.0:
			logger.warning(f"The column `{column}` had values greater than 1.0. It will be converted to a float between 0 and 1.")
			new_column = old_data[column] / 100
		else:
			new_column = old_data[column].astype(float)
		new_data[column] = new_column
	return new_data
def convert_string_to_number(value:str)->float:

	pattern = "[\d]+(?:[.][\d]+)?"
	if isinstance(value, str):
		match = re.search(pattern, value)
		try:
			result = float(match.group(0))
		except AttributeError:
			# No matches were found
			logger.error(f"Could not convert '{value}' to a number.")
			result = math.nan
		except ValueError:
			# The matched string could not be converted to a number
			logger.error(f"Could not convert '{value}' to a number.")
			result = math.nan
	else:
		result = value
	return result

def _fix_column_datatypes(table: pandas.DataFrame)->pandas.DataFrame:
	# Try to extract numerical data from the table.

	for column in table.columns:
		converted_column = table[column].apply(convert_string_to_number)
		table[column] = converted_column
	return table
def _parse_table(raw_table: pandas.DataFrame, key_column: str) -> Tuple[pandas.DataFrame, pandas.DataFrame]:
	"""
		Converts column headers to integers and moves all non-integer columns to a separate dataframe.
	"""
	# Make sure the column with the series names is the index of the table.

	raw_table = raw_table.sort_values(by = key_column)
	raw_table[key_column] = [str(i) for i in raw_table[key_column].tolist()]
	raw_table.set_index(key_column, inplace = True)
	# Extract the columns which indicate timepoints of observations. Should be integers.
	frequency_columns = get_numeric_columns(raw_table.columns)

	# Extract the columns with the trajectory identifiers and frequencies at each timepoint.
	time_table = raw_table[frequency_columns]
	# Convert the columns into integers. Makes them easier to work with if they have a standard raw_table type.
	converted_frequency_columns = [_convert_to_integer(i) for i in frequency_columns]
	time_table.columns = converted_frequency_columns
	# Sort the table columns
	time_table = time_table[sorted(time_table.columns)]

	try:
		# Test to see if the table was correctly converted to numerical datatypes.
		time_table[time_table.columns[0]].max()
	except TypeError:
		logger.warning(f"Some of the values could not be interpreted as numerical data. Attempting to extract numerical data from these values...")
		time_table = _fix_column_datatypes(time_table)

	# Drop any trajectories which are never detected.
	time_table = time_table[(time_table.T != 0).any()]

	# Make sure the table values are between 0 and 1 and are of type `float`
	time_table = _correct_math_scale(time_table)

	# Make sure the time table contains a column for timepoint `0`.
	if 0 not in time_table.columns:
		time_table[0] = 0.0  # Should be a float to match the dtype of the other columns
		logger.warning("Warning: The input table did not have values for timepoint 0. Adding 0% for each trajectory at timepoint 0")
	# Make sure there are no missing values.
	time_table = time_table.fillna(0)
	# Extract metadata for each series.
	info_table = raw_table[[i for i in raw_table.columns if i not in frequency_columns or i == key_column]]
	return time_table, info_table


def parse_genotype_table(filename: Path, sheet_name: str = 'Sheet1') -> Tuple[pandas.DataFrame, pandas.DataFrame]:
	""" Imports a table that lists pre-computed genotypes rather than trajectories."""
	data = import_table(filename, sheet_name = sheet_name)
	# For some reason some tables are annotated with 'genotype   ' with extra spaces.
	data.columns = [(i.strip() if isinstance(i, str) else i) for i in data.columns]
	if 'Genotype' in data.columns:
		key_column = 'Genotype'
	elif 'Unnamed: 0' in data.columns:
		key_column = 'Unnamed: 0'
	elif 'Trajectory' in data.columns:
		key_column = 'Trajectory'
	else:
		message = f"One of the columns needs to be labeled `Genotype`. Got {data.columns} instead from {filename}."
		raise ValueError(message)

	genotype_timeseries, genotype_info = _parse_table(data, key_column)
	# Make sure the genotype labels are prefixed with 'genotype-'
	if not re.search("genotype-[\d]+", genotype_timeseries.index[0]):
		genotype_timeseries.index.name = 'originalLabel'
		genotype_timeseries['Genotype'] = [f'genotype-{i}' for i in range(1, len(genotype_timeseries) + 1)]
		genotype_timeseries = genotype_timeseries.reset_index()
		genotype_timeseries = genotype_timeseries.set_index('Genotype')
		genotype_timeseries.pop('originalLabel')

	# Try to sort the genotypes by label, if posible.
	# Note: designed to sort labels of the form `genotype-[\d]+`
	try:
		sorted_index = sorted(genotype_timeseries.index, key = lambda s: float(s.split('-')[-1]))
	except ValueError:
		sorted_index = genotype_timeseries.index

	genotype_timeseries = genotype_timeseries.loc[sorted_index]

	# Remove extraneous whitespace.
	return genotype_timeseries, genotype_info


def parse_trajectory_table(filename: IOTYPE, sheet_name = 'Sheet1') -> Tuple[pandas.DataFrame, pandas.DataFrame]:
	"""
		Reads an excel or csv file. Assumes that the file has a `Trajectory` column and a column for each timepoint.
	Parameters
	----------
	filename: Path
		The table containing the trajectories and associated metadata. Can be an excel sheet or comma/tab delimited file.
	sheet_name: str; Default 'Sheet1'
		Indicates which sheet contains the data, if an excel table is given.
	Returns
	-------
	pandas.DataFrame, pandas.DataFrame
		A timeseries dataframe
			- Index -> str
				Names unique to each trajectory.
			- Columns -> int
				The timeseries points will correspond to the frequencies for each trajectory included with the input sheet.
				Each trajectory/timepoint will include the observed frequency at each timepoint.
		A metadata dataframe
			- Index -> str
				Identical index to the timeseries dataframe.
			- Columns -> str
				All columns from the original input table that do no correspond to timepoints.
	"""

	# Read in the data table.
	raw_data = import_table(filename, sheet_name)

	key_column = 'Trajectory'
	timeseries, info = _parse_table(raw_data, key_column)

	if 'genotype' in info:
		# This file was generated by a previous run.
		info.pop('genotype')
	return timeseries, info



from pathlib import Path

import matplotlib
import matplotlib.pyplot as plt
# plt.switch_backend('agg')
import pandas
import seaborn


def plot_heatmap(data: pandas.DataFrame, filename: Path):
	font = {
		'size': 20
	}
	use_annotations = len(data) < 20  # So the annotations are actually visible
	matplotlib.rc('font', **font)
	figsize = (20, 20)
	fig, ax = plt.subplots(figsize = figsize)
	ax.set_ylabel("Trajectory Label", size = 20)
	ax.set_xlabel("Trajectory Label", size = 20)
	ax.set_title("p-values of all mutational trajectories", size = 30)
	seaborn.heatmap(data, ax = ax, annot = use_annotations, cmap = 'Reds')
	plt.tight_layout()
	fig.savefig(str(filename), format = 'png')



import itertools
from typing import Dict, List, Tuple

import pandas

try:
	from muller.clustering.methods.difference import unlink_unrelated_trajectories
	from muller.clustering.metrics.pairwise_calculation_cache import PairwiseCalculationCache
	from muller.clustering.methods.population_genotypes_cache import PopulationGenotypes
except ModuleNotFoundError:
	from .difference import unlink_unrelated_trajectories
	from ..metrics.pairwise_calculation_cache import PairwiseCalculationCache
	from .population_genotypes_cache import PopulationGenotypes


def _group_trajectories_into_genotypes(pairs: Dict[Tuple[str, str], float], relative_cutoff: float, base_genotypes: List[List[str]] = None) -> List[
	List[str]]:
	"""
		Clusters all trajectories into related muller_genotypes.
		By default the first trajectory makes a genotype category
	Parameters
	----------
	pairs: PairwiseArrayType
		A dictionary mapping pairs to p-values.
	relative_cutoff: float; default 0.05
		The cutoff indicating two trajectories are related.
	base_genotypes: List[List[str]]
		A set of pre-defined genotypes.

	Returns
	-------
		A list of all muller_genotypes (each of which is a list of ints)
	"""
	if base_genotypes:
		genotype_candidates = base_genotypes
	elif pairs:
		genotype_candidates = [[min(pairs.keys())[0]]]  # by default the first trajectory forms the first genotype.
	else:
		genotype_candidates = []
	population_genotypes = PopulationGenotypes(genotype_candidates)
	seen = set(itertools.combinations(itertools.chain.from_iterable(genotype_candidates), 2))
	for (left, right), p_value in pairs.items():
		# ignore pairs that have already been sorted into a genotype.
		if (left, right) in seen or (right, left) in seen:
			continue
		seen.add((left, right))
		# are the genotypes related?
		if p_value > relative_cutoff:
			population_genotypes.merge_trajectories(left, right)
	genotype_candidates = population_genotypes.to_list()
	return genotype_candidates


def twostep_method(timeseries: pandas.DataFrame, pair_array: PairwiseCalculationCache, similarity_breakpoint: float, difference_breakpoint: float,
		starting_genotypes: List[List[str]]) -> List[List[str]]:
	"""
		Clusters trajectories into muller_genotypes.
	Parameters
	----------
	timeseries: pandas.DataFrame
		The timeseries output from import_timeseries.
			- Index -> str
				Names unique to each trajectory.
			- Columns -> int
				The timeseries points will correspond to the frequencies for each trajectory included with the input sheet.
				Each trajectory/timepoint will include the observed frequency at each timepoint.
	pair_array: PairwiseArrayType
	similarity_breakpoint: float
	difference_breakpoint: float
	starting_genotypes: List[List[str]]
		A list of trajectories to use as the initial genotypes. Each sub-list represents a single genotype.
	Returns
	-------
	List[List[str]]
		A list of the genotypes, where each genotype is itself a list of the name of each member trajectory.
		Ex. [
			[A1, A2, A3],
			[B1, B2],
			[C1]
		]
	"""

	# Trajectories represent the population frequencies at each timepoint
	# Each row represents a single timepoint, each column represents a mutation.
	numerical_array = pair_array.asdict()
	population_genotypes = _group_trajectories_into_genotypes(numerical_array, similarity_breakpoint, starting_genotypes)

	# at the end, look at all trajectories that are not listed and
	# append them as their own category.
	# List of all trajectories
	flattened_genotypes = list(itertools.chain.from_iterable(population_genotypes))
	# Any missing trajectories. Will probably be empty
	other_trajectories = [i for i in timeseries.index if i not in flattened_genotypes]
	population_genotypes.append(other_trajectories)

	# finally, for each genotype, make sure each trajectory pair has some
	# non-trivial linkage (say, >0.0005). this avoids falsely linking together
	# trajectories. if not, divide the offending trajectories into two
	# camps, and sort the rest according to which one they are more
	# closely linked to. repeat until everything is linked.
	# pprint(pair_array)

	while True:
		starting_size_of_the_genotype_array = len(population_genotypes)
		population_genotypes = unlink_unrelated_trajectories(population_genotypes[:], numerical_array, difference_breakpoint)
		if len(population_genotypes) == starting_size_of_the_genotype_array:
			break

	return [i for i in population_genotypes if i]  # Only return non-empty lists.



import matplotlib.pyplot as plt
from matplotlib.figure import Axes  # For autocomplete

plt.switch_backend('agg')
import pandas
from pathlib import Path
from typing import Dict, Optional

try:
	from muller.widgets import get_numeric_columns
except ModuleNotFoundError:
	from ..widgets import get_numeric_columns
BASE_COLOR = "#333333"


def plot_timeseries(timeseries: pandas.DataFrame, palette: Dict[str, str], ax: Optional[Axes] = None, filename: Optional[Path] = None):
	if ax is None:
		fig, ax = plt.subplots(figsize = (10, 5))
	ax.set_ylabel('frequency')
	plot_title = 'Genotypes' if 'genotype' in timeseries.index[0] else 'Trajectories'
	ax.set_title(plot_title)
	ax.set_ylim(0, 1.01)

	numeric_columns = list(get_numeric_columns(timeseries.columns))

	for trajectory_id, trajectory in timeseries.iterrows():
		color = palette.get(trajectory_id, BASE_COLOR)
		trajectory_timeseries = sorted((column, trajectory[column]) for column in numeric_columns)
		x_values, y_values = zip(*trajectory_timeseries)
		ax.plot(x_values, y_values, '-o', color = color, label = trajectory_id, markersize = 2)
	if filename:
		plt.tight_layout()
		plt.savefig(str(filename), dpi = 200)
	return ax


# noinspection PyTypeChecker
def plot_genotypes(timeseries: pandas.DataFrame, mutational_genotypes: pandas.DataFrame, filename: Path, genotype_colors: Dict[str, str],
		trajectory_palette: Dict[str, str]):
	"""
		Plots the clustered muller_genotypes.
	Parameters
	----------
	timeseries: pandas.DataFrame
		The table with each trajectory used to generate the muller_genotypes.
	filename: Optional[Path]
		Path to save the genotype plot.
	mutational_genotypes: pandas.DataFrame
	genotype_colors: Dict[str,str]
		A mapping of muller_genotypes to their corresponding colors.
	trajectory_palette: Dict[str,str]
		Maps trajectories to their respective genotype color.
	Returns
	-------

	"""
	if 'members' in mutational_genotypes:
		mutational_genotypes.pop('members')

	# Plot all trajectories
	grid = plt.GridSpec(2, 3, wspace = 0.4, hspace = 0.3)
	if timeseries is not None:
		taxes: Axes = plt.subplot(grid[0, 0:2])
		plot_timeseries(timeseries, trajectory_palette, taxes)

	# Plot clustered muller_genotypes. Should be same as above, but colored based on genotype cluster.
	# Plot the mean of each cluster
	if timeseries is not None:
		gaxes: Axes = plt.subplot(grid[1, :])
	else:
		gaxes: Axes = plt.subplot(grid[:, :])

	plot_timeseries(mutational_genotypes, genotype_colors, gaxes)
	# noinspection PyUnboundLocalVariable
	if len(mutational_genotypes) > 20:
		bbox = (1, 1.8)
	else:
		bbox = (1, 2)
	legend_font_properties = {'size': 5}
	plt.legend(
		loc = 'right',
		ncol = 2,
		bbox_to_anchor = bbox,
		prop = legend_font_properties,
		title = 'Genotypes',
	)

	if filename:
		plt.savefig(str(filename), dpi = 500, format = 'png')
	else:
		plt.show()



import itertools
from typing import Dict, List, Tuple

import pandas


# noinspection PyUnresolvedReferences,PyTypeChecker
def _find_weakest_pair(trajectories: pandas.Series, link_cut: float) -> Tuple[str, str]:
	return (trajectories - link_cut).abs().idxmin()


# noinspection PyTypeChecker
def _divide_genotype(genotype: List[str], unlinked_trajectories: pandas.Series, link_cut: float) -> Tuple[List[str], List[str]]:
	"""
		Splits a genotype into smaller muller_genotypes if some members are not related to some other members. This may happen
		when a member is included into a genotype due to its paired member but ends up not related to some of
		the other members that where found later.
	Parameters
	----------
	genotype: Genotype
		The genotype to split.
	unlinked_trajectories: pandas.Series
		A series of trajectory pairs and their corresponding p-value.
		Index:
			- left
			- right
		Values:
			- pvalue
	link_cut: float
		The cuttoff value to choose whether a member is related to the other members.

	Returns
	-------
		a tuple of two muller_genotypes.
	"""
	# Find the index of the minimum p-value after subtracting the link cutoff.
	# Form two new muller_genotypes based on the two trajectories corresponding to the minimum p-value
	# noinspection PyUnresolvedReferences
	# new_genotype_1_base, new_genotype_2_base = (unlinked_trajectories - link_cut).abs().idxmin()
	new_genotype_1_base, new_genotype_2_base = _find_weakest_pair(unlinked_trajectories, link_cut)
	# Get the row with the identified minimum mp-value.

	# Make sure genotype 1 includes the lower id-value. Not important, but maintains parity with matlab script.
	new_genotype_1_base, new_genotype_2_base = sorted([new_genotype_1_base, new_genotype_2_base])
	new_genotype_1 = [new_genotype_1_base]
	new_genotype_2 = [new_genotype_2_base]

	# Sort each member in the current genotype into one of the new muller_genotypes.
	for genotype_member in genotype:
		# Check if the current genotype member is already contained in one of the muller_genotypes.
		# Should only be one of the two trajectories used to form a new genotype.
		if not (genotype_member in new_genotype_1 or genotype_member in new_genotype_2):
			# Use the highest p-value to determine which genotype to add the member to.
			# P-values should correspond to the current member and the base member of the new muller_genotypes.
			p_value_1 = unlinked_trajectories.loc[new_genotype_1_base, genotype_member]
			p_value_2 = unlinked_trajectories.loc[new_genotype_2_base, genotype_member]

			if p_value_1 >= p_value_2:
				new_genotype_1.append(genotype_member)
			else:
				new_genotype_2.append(genotype_member)
	return new_genotype_1, new_genotype_2


def unlink_unrelated_trajectories(all_genotypes: List[List[str]], pair_array: Dict[Tuple[str, str], float], link_cutoff: float) -> List[List[str]]:
	"""
		Splits each genotype if any of its members are not related enough to the other members. Genotypes will continue
		to be split until the p-values for all pairwise members are beneath the cutoff.
	Parameters
	----------
	all_genotypes: List[Genotype]
		A list of all muller_genotypes.
	pair_array: PairWiseArrayType
		A mapping of all pairwise p-values.
	link_cutoff: float
		The cuttoff value to determine if a given pair of trajectories is unrelated.

	Returns
	-------
		A list of the new muller_genotypes.

	"""
	for genotype in all_genotypes:
		if len(genotype) > 1:
			# Iterate over all possible pairs of genotype members.
			combination_pairs = [(left, right, pair_array[left, right]) for left, right in itertools.permutations(genotype, 2)]
			# Combine all pairs and p-values into a dataframe for convienience.
			genotype_combinations: pandas.DataFrame = pandas.DataFrame(combination_pairs, columns = ['left', 'right', 'pvalue'])
			genotype_combinations: pandas.Series = genotype_combinations.set_index(['left', 'right'])['pvalue']
			# Get a dataframe of all trajectories in this genotype which are significantly different than the
			# current pair of trajectories.
			# noinspection PyTypeChecker
			unlinked_trajectories = genotype_combinations[genotype_combinations < link_cutoff].dropna()

			if len(unlinked_trajectories) != 0:
				# Split the current genotype into two smaller but more internally-related all_genotypes.
				new_genotype_1, new_genotype_2 = _divide_genotype(genotype[:], genotype_combinations, link_cutoff)

				all_genotypes.append(new_genotype_1)
				all_genotypes.append(new_genotype_2)
				all_genotypes.remove(genotype)
	# noinspection PyTypeChecker
	return sorted(all_genotypes, key = len)



from pathlib import Path
from typing import Dict, List, Optional

import pandas
import pygraphviz

from muller import widgets


def get_node_label_properties(identity: str, genotype_color: str, annotation: List[str]) -> Dict[str, str]:
	luminance = widgets.calculate_luminance(genotype_color)
	if luminance > 0.5:
		font_color = "#333333"
	else:
		font_color = "#FFFFFF"

	label = [identity] + annotation
	label = "\n".join(label)

	label_properties = {
		'label':     label,
		'fontcolor': font_color
	}
	return label_properties


def flowchart(edges: pandas.DataFrame, palette: Dict[str, str], annotations: Dict[str, List[str]] = None, filename: Optional[Path] = None):
	"""
		Creates a lineage plot showing the ancestry of each genotype.
	Parameters
	----------
	edges: pandas.DataFrame
		Should have three columns: 'parent', 'identity', 'score'
	palette
	annotations
	filename

	Returns
	-------

	"""
	if annotations is None:
		annotations = {}
	graph = pygraphviz.AGraph(splines = 'ortho')
	graph.node_attr['shape'] = 'box'
	graph.node_attr['style'] = 'filled,rounded'
	graph.node_attr['fontname'] = 'lato'
	graph.edge_attr['dir'] = 'forward'

	for identity in edges['identity'].unique():
		genotype_color = palette.get(identity)
		graph.add_node(identity, color = "#333333", fillcolor = genotype_color)
		node = graph.get_node(identity)
		node_label_properties = get_node_label_properties(identity, genotype_color, annotations.get(identity, []))
		node.attr.update(node_label_properties)

	for index, row in edges.iterrows():
		parent = row['parent']
		identity = row['identity']
		score = row['score']
		#if score < 0: score = 0
		graph.add_edge(parent, identity, tooltip = f"parent", headlabel = f"{score:.1f}", labeldistance = 2.0)
	if filename:
		graph.draw(str(filename), prog = 'dot')
	return graph
